import struct
import json
from unittest import TestCase
from hbase.admin import HBaseAdmin
from hbase.rest_client import HBaseRESTClient
from hbase.put import Put
from hbase.scan import Scan
from hbase.scan_filter_helper import (
    build_base_scanner,
    build_prefix_filter,
    build_row_filter,
    build_value_filter,
    build_single_column_value_filter,
)
from hbase.utils import result_parser


class TestScan(TestCase):
    @classmethod
    def setUpClass(cls):
        client = HBaseRESTClient(hosts_list=["http://localhost:8080"])
        admin = HBaseAdmin(client)
        put = Put(client=client)

        admin.table_create_or_update("test_scan", [{"name": "cf"}])
        for i in range(1, 101):
            put.put(
                tbl_name="test_scan",
                row_key=f"row-{i}",
                cell_map={
                    "cf:age": i,
                    "cf:email": f"ex{i}@grr.la",
                    "cf:text": "this is just text test",
                },
            )
        client.session.close()

    @classmethod
    def tearDownClass(cls):
        client = HBaseRESTClient(hosts_list=["http://localhost:8080"])
        admin = HBaseAdmin(client)
        admin.table_delete("test_scan")
        client.session.close()

    def setUp(self):
        self.client = HBaseRESTClient(hosts_list=["http://localhost:8080"])
        self.admin = HBaseAdmin(client=self.client)
        self.put = Put(client=self.client)
        self.scan = Scan(client=self.client)

    def tearDown(self):
        self.client.session.close()

    def test_scan_default(self):
        next, data = self.scan.scan("test_scan")
        self.assertEqual(len(data["row"]), 100)
        while next:
            next, data = self.scan.scan_next()
            assert next is False
            assert data is None
        assert self.scan.scanner is None

    def test_scan_with_params(self):
        scanner_def = build_base_scanner(
            startRow="row-9", endRow="row-99", column=["cf:email"]
        )
        _, data = self.scan.scan(tbl_name="test_scan", scanner_payload=scanner_def)
        self.assertEqual(len(data["row"]), 10)
        self.assertEqual(self.scan.delete_scanner(), 200)

    def test_scan_prefix_filter(self):
        filter = build_prefix_filter(row_perfix="row-8", column=["cf:email"])
        _, data = self.scan.scan(tbl_name="test_scan", scanner_payload=filter)
        self.assertEqual(len(data["row"]), 11)
        self.assertEqual(self.scan.delete_scanner(), 200)

    def test_scan_row_filter(self):
        filter = build_row_filter(
            row_value="row-1",
            operation="EQUAL",
        )
        _, data = self.scan.scan(tbl_name="test_scan", scanner_payload=filter)
        rows = data["row"]
        self.assertEqual(len(rows), 1)
        self.assertEqual(self.scan.delete_scanner(), 200)

    def test_scan_value_filter(self):
        filter = build_value_filter(
            value=10,
            operation="LESS",
        )
        _, data = self.scan.scan(tbl_name="test_scan", scanner_payload=filter)
        rows = data["row"]
        values = []
        for r in rows:
            values.append(struct.unpack(">q", r["cell"][0]["$"])[0])
        for v in values:
            assert v < 10
        self.assertEqual(self.scan.delete_scanner(), 200)

    def test_singe_column_value_filter(self):
        filter = build_single_column_value_filter(
            family="cf", qualifier="age", value=10, operation="LESS"
        )
        _, data = self.scan.scan(tbl_name="test_scan", scanner_payload=filter)
        rows = data["row"]
        self.assertEqual(len(rows), 9)
        values = []
        for r in rows:
            values.append(struct.unpack(">q", r["cell"][0]["$"])[0])
        for v in values:
            assert v < 10
        self.assertEqual(self.scan.delete_scanner(), 200)

    def test_table_scan(self):
        url_sufix = "/test_scan/*?column=cf:email&startrow=row-99&endrow=row-9&reversed=true"
        data = self.client.send_request(
            method="GET",
            url_suffix=url_sufix
        )
        result = result_parser(json.loads(data))
        self.assertEqual(result['row'][0]['key'], b'row-99')

    def test_value_filter_substring(self):
        filter = build_value_filter(
            value="this", operation="EQUAL", comparator="substring"
        )
        _, result = self.scan.scan(tbl_name="test_scan", scanner_payload=filter)
        self.assertEqual(len(result["row"]), 100)



class Delete(object):
    def __init__(self, client):
        self.client = client

    @staticmethod
    def _build_url_suffix(tbl_name, row_key, column_family=None):
        if column_family is not None:
            return f"/{tbl_name}/{row_key}/{column_family}"
        return f"/{tbl_name}/{row_key}"

    def delete(self, tbl_name, row_key, column_family=None):
        return self.client.send_request(
            method="DELETE",
            url_suffix=self._build_url_suffix(tbl_name, row_key, column_family),
        )



import json
from unittest import TestCase, skip
from base64 import b64decode, b64encode
from time import time
import struct
import pickle
from hbase.utils import b64_encoder
from hbase.admin import HBaseAdmin
from hbase.rest_client import HBaseRESTClient
from hbase.put import Put
from hbase.get import Get
from hbase.delete import Delete


class TestPut(TestCase):
    def setUp(self):
        self.client = HBaseRESTClient(hosts_list=["http://localhost:8080"])
        self.admin = HBaseAdmin(client=self.client)
        self.put = Put(client=self.client)
        self.get = Get(client=self.client)
        self.delete = Delete(client=self.client)

    def tearDown(self):
        self.client.session.close()

    def test_put(self):
        self.admin.table_create_or_update(
            table_name="test_tbl", params_list=[{"name": "cf"}]
        )

        self.put.put(
            tbl_name="test_tbl",
            row_key="samir@example.com",
            cell_map={"cf:amount": 250.511234},
        )

        self.put.put("test_tbl", "samir@example.com", {"cf:gender": "F"})
        self.put.put(
            "test_tbl", "samir@example.com", {"cf:countries": ["US", "BA", "DE", "SE"]}
        )
        self.put.put("test_tbl", "samir@example.com", {"cf:id": 13455})

        # Test int value put/get
        cf_id = self.get.get("test_tbl", "samir@example.com", "cf:id")
        byte_value = cf_id.get("row")[0]["cell"][0]["$"]
        assert struct.unpack(">q", byte_value)[0] == 13455

        # Test float value put/get
        cf_amount = self.get.get("test_tbl", "samir@example.com", "cf:amount")
        byte_value = cf_amount.get("row")[0]["cell"][0]["$"]
        assert struct.unpack(">d", byte_value)[0] == 250.511234

        # Test string value put/get
        cf_gender = self.get.get("test_tbl", "samir@example.com", "cf:gender")
        assert cf_gender.get("row")[0]["cell"][0]["$"].decode("utf-8") == "F"

        # Test python complex type: list
        cf_countries = self.get.get("test_tbl", "samir@example.com", "cf:countries")
        byte_value = cf_countries.get("row")[0]["cell"][0]["$"]
        assert pickle.loads(byte_value) == ["US", "BA", "DE", "SE"]

        # Test multiple cf put
        self.put.put(
            tbl_name="test_tbl",
            row_key="test@example.com",
            cell_map={
                "cf:gender": "M",
                "cf:amount": 34.567,
                "cf:countries": ["US", "BA", "DE", "SE"],
                "cf:id": 4567890,
            },
        )

        # Test int value put/get
        cf_id = self.get.get("test_tbl", "test@example.com", "cf:id")
        byte_value = cf_id.get("row")[0]["cell"][0]["$"]
        assert struct.unpack(">q", byte_value)[0] == 4567890

        # Test float value put/get
        cf_amount = self.get.get("test_tbl", "test@example.com", "cf:amount")
        byte_value = cf_amount.get("row")[0]["cell"][0]["$"]
        assert struct.unpack(">d", byte_value)[0] == 34.567

        # Test string value put/get
        cf_gender = self.get.get("test_tbl", "test@example.com", "cf:gender")
        assert cf_gender.get("row")[0]["cell"][0]["$"].decode("utf-8") == "M"

        # Test python complex type: list
        cf_countries = self.get.get("test_tbl", "test@example.com", "cf:countries")
        byte_value = cf_countries.get("row")[0]["cell"][0]["$"]
        assert pickle.loads(byte_value) == ["US", "BA", "DE", "SE"]

        # Test put with explicit timestamp
        status = self.put.put(
            tbl_name="test_tbl",
            row_key="timestamp_test",
            cell_map={
                "cf:gender": "M",
                "cf:amount": 34.567,
                "cf:countries": ["US", "BA", "DE", "SE"],
                "cf:id": 4567890,
            },
            timestamp=5555,
        )
        data = self.get.get("test_tbl", "timestamp_test")
        timestamp = data["row"][0]["cell"][0]["timestamp"]
        self.assertEqual(timestamp, 5555)

    def test_bulk_put_raw(self):
        p = {
            "Row": [
                {
                    "key": b64_encoder("timestamp6"),
                    "Cell": [
                        {
                            "column": b64_encoder("cf:t"),
                            "timestamp": round(time() * 1000),
                            "$": b64_encoder("a6"),
                        }
                    ],
                },
                {
                    "key": b64_encoder("timestamp7"),
                    "Cell": [
                        {
                            "column": b64_encoder("cf:t"),
                            "timestamp": round(time() * 1000),
                            "$": b64_encoder("a6"),
                        }
                    ],
                },
            ]
        }
        self.client.send_request(
            method="PUT",
            url_suffix="/test_tbl/fakekey",
            payload=p,
        )

    def test_build_bulk_payload(self):
        test_data = [("timestamp77", {"cf:col1": "value1"})]
        paylod = self.put._build_bulk_payload(test_data)
        paylod3 = self.put._build_payload(
            row_key="timestamp77", cell_map={"cf:col1": "value1"}
        )
        self.assertEqual(paylod, paylod3)
        self.assertTrue(isinstance(paylod["Row"], list))
        self.assertTrue(isinstance(json.dumps(paylod), str))

    def test_bulk_put(self):
        self.admin.table_create_or_update(
            table_name="test_tbl", params_list=[{"name": "cf"}]
        )
        bulk_payload = [
            ("row_key1", {"cf:col1": "value1", "cf:col2": "value2"}),
            ("row_key2", {"cf:col1": "value1", "cf:col2": "value2"}),
        ]
        self.put.bulk("test_tbl", rows=bulk_payload, timestamp=round(time() * 1000))
        data = self.get.get("test_tbl", "row_key2", "cf:col1")
        self.assertEqual(data["row"][0]["cell"][0]["$"], b"value1")

        bulk_payload2 = [
            ("test_bulk1", {"cf:age": 28, "cf:gender": "M", "cf:amount": 2500.45}),
            ("test_bulk2", {"cf:age": 29, "cf:gender": "F", "cf:amount": 2501.45}),
            ("test_bulk3", {"cf:age": 30, "cf:gender": "M", "cf:amount": 2502.45}),
            ("test_bulk4", {"cf:age": 30, "cf:gender": "M", "cf:amount": 2502.45}),
            ("test_bulk5", {"cf:age": 30, "cf:gender": "M", "cf:amount": 2502.45}),
            ("test_bulk6", {"cf:age": 30, "cf:gender": "M", "cf:amount": 2502.45}),
            ("test_bulk7", {"cf:age": 30, "cf:gender": "M", "cf:amount": 2502.45}),
            ("test_bulk8", {"cf:age": 30, "cf:gender": "M", "cf:amount": 2502.45}),
            ("test_bulk9", {"cf:age": 30, "cf:gender": "M", "cf:amount": 2502.45}),
            ("test_bulk10", {"cf:age": 30, "cf:gender": "M", "cf:amount": 2502.45}),
        ]
        self.put.bulk("test_tbl", rows=bulk_payload2)
        data = self.get.get("test_tbl", "test_bulk7", "cf:gender")
        self.assertEqual(data["row"][0]["cell"][0]["$"], b"M")



import json
from unittest import TestCase, skip
from base64 import b64decode, b64encode
from time import time
import pickle
from hbase.utils import b64_encoder
from hbase.admin import HBaseAdmin
from hbase.rest_client import HBaseRESTClient


class TestAdmin(TestCase):
    def setUp(self):
        self.client = HBaseRESTClient(hosts_list=["http://localhost:8080"])
        self.admin = HBaseAdmin(client=self.client)

    def tearDown(self):
        self.client.session.close()

    def test_cluster_status(self):
        result = self.admin.cluster_status()
        assert isinstance(result, str)
        result_dict = json.loads(result)
        assert "LiveNodes" in result
        assert isinstance(result_dict.get("LiveNodes"), list)

    def test_cluster_version(self):
        result = self.admin.cluster_version()
        assert "Version" in json.loads(result)

    def test_namespaces(self):
        result = self.admin.namespaces()
        result_dict = json.loads(result)
        assert "Namespace" in json.loads(result)
        assert "default" in result_dict.get("Namespace")

    def test_namespace(self):
        result = self.admin.namespace(namespace_name="default")
        assert isinstance(json.loads(result), dict)

    def test_create_namespace(self):
        self.admin.namespace_create(namespace_name="new_namespace3")
        result = self.admin.namespaces()
        result_dict = json.loads(result)
        assert "new_namespace3" in result_dict.get("Namespace")

    def test_namespace_delete(self):
        self.admin.namespace_create(namespace_name="new_namespace3")
        self.admin.namespace_delete(namespace_name="new_namespace3")
        result = self.admin.namespaces()
        result_dict = json.loads(result)
        assert "new_namespace3" not in result_dict.get("Namespace")

    def test_tables(self):
        result = self.admin.tables()
        result_dict = json.loads(result)
        assert "table" in result_dict
        assert isinstance(result_dict.get("table"), list)

    def test_table_schema(self):
        self.admin.table_create_or_update(
            table_name="test_tbl", params_list=[{"name": "cfTest"}]
        )
        result = self.admin.table_schema(table_name="test_tbl")
        result_dict = json.loads(result)
        self.assertEqual(result_dict.get("name"), "test_tbl")
        self.assertEqual(result_dict.get("ColumnSchema")[0]["name"], "cfTest")

    def test_table_create_or_update(self):
        self.admin.table_create_or_update(
            table_name="test_tbl",
            params_list=[{"name": "cfTest", "BLOCKSIZE": "65537"}],
        )
        result = self.admin.table_schema(table_name="test_tbl")
        result_dict = json.loads(result)
        self.assertEqual(result_dict.get("name"), "test_tbl")
        self.assertEqual(result_dict.get("ColumnSchema")[0]["BLOCKSIZE"], "65537")

    def test_table_regions(self):
        self.admin.table_create_or_update(
            table_name="test_tbl",
            params_list=[{"name": "cfTest", "BLOCKSIZE": "65537"}],
        )
        result = self.admin.table_regions(table_name="test_tbl")
        result_dict = json.loads(result)
        assert result_dict["name"] == "test_tbl"
        assert isinstance(result_dict["Region"], list)

    def test_table_delete(self):
        self.admin.table_create_or_update(
            table_name="test_tbl2",
            params_list=[{"name": "cfTest", "BLOCKSIZE": "65537"}],
        )
        self.admin.table_delete(table_name="test_tbl2")
        result = self.admin.tables()
        result_dict = json.loads(result)
        tbls_list = [a["name"] for a in result_dict["table"]]
        assert "test_tbl2" not in tbls_list



import json
from hbase.utils import result_parser


class Scan(object):
    def __init__(self, client):
        self.client = client
        self.scanner = None

    @staticmethod
    def _build_url_suffix(tbl_name):
        return f"/{tbl_name}/scanner"

    def _get_scanner(self, tbl_name, scanner_payload='<Scanner batch="1000"/>'):
        hbase_host = self.client.get_hbase_host()
        url_suffix = self._build_url_suffix(tbl_name)
        response = self.client.session.put(
            url=f"{hbase_host}{url_suffix}",
            data=scanner_payload,
            headers={"Content-Type": "text/xml", "accept": "text/xml"},
        )
        if response.status_code == 201:
            self.scanner = response.headers.get("Location")
            return self.scanner
        return {"error": response.content, "status_code": response.status_code}

    def scan_next(self):
        response = self.client.session.get(
            url=self.scanner,
            headers=self.client.headers,
        )
        if response.status_code == 204:
            self.delete_scanner()
            self.scanner = None
            return False, None
        return (
            True,
            result_parser(response.json()) if response.content is not None else None,
        )

    def scan(self, tbl_name, scanner_payload='<Scanner batch="1000"/>'):
        scan_url = None
        if self.scanner is None:
            scan_url = self._get_scanner(tbl_name, scanner_payload)
        if isinstance(scan_url, dict):
            return scan_url
        return self.scan_next()

    def delete_scanner(self):
        response = self.client.session.delete(
            url=self.scanner, headers=self.client.headers
        )
        return response.status_code



from unittest import TestCase
from hbase.utils import b64_encoder, result_parser
import struct
from base64 import b64encode, b64decode


class TestUtils(TestCase):
    def test_b64_encoder(self):
        test_values = ["a", b"test", 10, 250.511234, [1, 3, 5]]
        for value in test_values:
            assert isinstance(b64_encoder(value), str)

    def test_result_parser(self):
        test_payload = {
            "Row": [
                {
                    "key": "c2FtaXJAZXhhbXBsZS5jb20=",
                    "Cell": [
                        {
                            "column": "Y2Y6YW1vdW50",
                            "timestamp": 1636391507135,
                            "$": "QG9QXAdn004=",
                        }
                    ],
                }
            ]
        }
        parsed = result_parser(test_payload)
        # {'row': [{'key': b'samir@example.com', 'cell': [{'column': b'cf:amount', 'timestamp': 1636391507135, '$': b'@oP\\\x07g\xd3N'}]}]}
        self.assertTrue("row" in parsed.keys())
        self.assertTrue(isinstance(parsed["row"], list))
        self.assertTrue("key" in parsed["row"][0].keys())
        self.assertTrue("cell" in parsed["row"][0].keys())
        self.assertEqual(parsed["row"][0]["key"].decode("utf-8"), "samir@example.com")
        self.assertEqual(
            parsed["row"][0]["cell"][0]["column"].decode("utf-8"), "cf:amount"
        )
        self.assertTrue(isinstance(parsed["row"][0]["cell"][0]["$"], bytes))



import string
import time
from queue import Queue
import random
from hbase.rest_client import HBaseRESTClient
from hbase.admin import HBaseAdmin
from hbase.put import Put
from hbase.scan import Scan
import logging
import threading
from threading import Thread


logging.basicConfig(level=logging.INFO)
times = []


def queue_put_payload(queue, payload):
    queue.put(payload)


def buid_key():
    return "row-" + "".join(random.choices(string.ascii_letters + string.digits, k=6))


def get_random_county():
    return "".join(
        random.choices(
            [
                "US",
                "BA",
                "DE",
                "SE",
                "FR",
                "IN",
                "NO",
                "SK",
                "AU",
                "UK",
                "DK",
                "NL",
                "BE",
                "ES",
            ]
        )
    )


def get_random_int():
    return random.randint(1, 1000)


def buld_put_payload():
    return {
        "tbl": "tbl_int",
        "row-key": buid_key(),
        "cf_value_map": {
            "cf:country": get_random_county(),
            "cf:count": get_random_int(),
        },
    }


def send_put(queue, total, put_instance):
    while True:
        try:
            progress = queue.qsize()
            if progress == 0:
                break
            percents_done = (1.0 - (progress) / total) * 100
            if progress % 10000 == 0:
                logging.info(
                    " %s - %s  [%s]%% "
                    % (threading.current_thread().name, progress, round(percents_done))
                )
            put_payload = queue.get()
            start = time.time()
            put_instance.put(
                tbl_name=put_payload["tbl"],
                row_key=put_payload["row-key"],
                cell_map=put_payload["cf_value_map"],
            )
            end = time.time()
            times.append(end - start)
            queue.task_done()
        except Exception as e:
            logging.exception("Error updating releases_releaseartistrole table")
            queue.task_done()


if __name__ == "__main__":
    NUMBER_OF_THREDDS = 1
    WRITE_ROWS = 20000
    client = HBaseRESTClient(hosts_list=["http://localhost:8080"])
    scan = Scan(client=client)
    admin = HBaseAdmin(client=client)
    admin.table_create_or_update(table_name="tbl_int", params_list=[{"name": "cf"}])
    put = Put(client=client)
    q = Queue()
    for i in range(1, WRITE_ROWS):
        put_payload = buld_put_payload()
        queue_put_payload(q, put_payload)
    logging.info(f"Starting {NUMBER_OF_THREDDS} writer threads")
    for t in range(0, NUMBER_OF_THREDDS):
        worker = Thread(target=send_put, args=(q, WRITE_ROWS, put))
        worker.setDaemon(True)
        worker.start()
    q.join()

    avg_request_time = sum(times) / WRITE_ROWS
    print(f"Max request time [ms] {max(times)*1000}")
    print(f"Min request time [ms] {min(times) *1000}")
    print(f"Average put op time [ms] {avg_request_time *1000}")



import time
from unittest import TestCase
from hbase.admin import HBaseAdmin
from hbase.rest_client import HBaseRESTClient
from hbase.put import Put
from hbase.get import Get
from hbase.utils import Bytes


class TestGet(TestCase):
    @classmethod
    def setUpClass(cls):
        client = HBaseRESTClient(hosts_list=["http://localhost:8080"])
        admin = HBaseAdmin(client)
        put = Put(client=client)

        admin.table_create_or_update("test_get", [{"name": "cf", "VERSIONS": 3}])
        for i in range(1, 101):
            put.put(
                tbl_name="test_get",
                row_key=f"row-{i}",
                cell_map={"cf:age": i, "cf:amount": 34.56, "cf:name": "John"},
            )
        client.session.close()

    @classmethod
    def tearDownClass(cls):
        client = HBaseRESTClient(hosts_list=["http://localhost:8080"])
        admin = HBaseAdmin(client)
        admin.table_delete("test_get")
        client.session.close()

    def setUp(self):
        self.client = HBaseRESTClient(hosts_list=["http://localhost:8080"])
        self.admin = HBaseAdmin(client=self.client)
        self.put = Put(client=self.client)
        self.get = Get(client=self.client)

    def tearDown(self):
        self.client.session.close()

    def test_get(self):
        result = self.get.get(
            tbl_name="test_get",
            row_key="row-5",
        )
        value_int = Bytes.to_int(result["row"][0]["cell"][0]["$"])
        self.assertEqual(value_int, 5)
        value_float = Bytes.to_float(result["row"][0]["cell"][1]["$"])
        self.assertEqual(value_float, 34.56)
        string_value = Bytes.to_string(result["row"][0]["cell"][2]["$"])
        self.assertEqual(string_value, "John")

        # Add more recent version for "row-1"
        self.put.put(
            tbl_name="test_get",
            row_key="row-1",
            cell_map={"cf:age": 50, "cf:amount": 50.1, "cf:name": "JohnNew"},
        )
        timestamp = round(time.time() * 1000) - 200
        result_2_versions = self.get.get(
            "test_get", "row-1", "cf:name", number_of_versions=3
        )
        result_timestamp = self.get.get(
            "test_get", "row-1", "cf:name", timestamp=timestamp
        )
        name = Bytes.to_string(result_2_versions["row"][0]["cell"][0]["$"])
        self.assertEqual(name, "JohnNew")
        self.assertEqual(
            Bytes.to_string(result_timestamp["row"][0]["cell"][0]["$"]), "John"
        )



import json
import responses
from unittest import TestCase
from mock import patch
from hbase.rest_client import HBaseRESTClient


class TestRESTClient(TestCase):
    def setUp(self):
        self.client = HBaseRESTClient(hosts_list=["http://localhost:8080"])

    def tearDown(self):
        self.client.session.close()

    def test_get_hbase_host(self):
        assert self.client.get_hbase_host() == "http://localhost:8080"

    @patch("requests.Session.get")
    def test_GET_method(self, mocked_fnc):
        self.client.send_request("GET", "/test")
        mocked_fnc.assert_called_with(
            url="http://localhost:8080/test",
            headers={"accept": "application/json", "content-type": "application/json"},
            timeout=10,
        )

    @patch("requests.Session.post")
    def test_POST_method(self, mocked_fnc):
        self.client.send_request("POST", "/test", {"a": 1})
        mocked_fnc.assert_called_with(
            url="http://localhost:8080/test",
            data='{"a": 1}',
            headers={"accept": "application/json", "content-type": "application/json"},
            timeout=10,
        )

    @patch("requests.Session.put")
    def test_PUT_method(self, mocked_fnc):
        self.client.send_request("PUT", "/test", {"a": 1})
        mocked_fnc.assert_called_with(
            url="http://localhost:8080/test",
            data='{"a": 1}',
            headers={"accept": "application/json", "content-type": "application/json"},
            timeout=10,
        )

    @patch("requests.Session.delete")
    def test_DELETE_method(self, mocked_fnc):
        self.client.send_request("DELETE", "/test")
        mocked_fnc.assert_called_with(
            url="http://localhost:8080/test",
            headers={"accept": "application/json", "content-type": "application/json"},
            timeout=10,
        )

    @responses.activate
    def test_send_request_success(self):
        expected = {"table": [{"name": "message"}, {"name": "new_table"}]}
        responses.add(
            responses.GET,
            "http://localhost:8080/namespaces/default/tables",
            json=expected,
            status=200,
        )
        result = self.client.send_request(
            method="GET", url_suffix="/namespaces/default/tables"
        )
        assert json.loads(result) == expected

    @responses.activate
    def test_send_request_failedd(self):
        payload = "Bad request"
        responses.add(
            responses.GET,
            "http://localhost:8080/namespaces/default/tables",
            json=payload,
            status=400,
        )
        result = self.client.send_request(
            method="GET", url_suffix="/namespaces/default/tables"
        )
        assert result == {"error": '"Bad request"', "status_code": 400}



import json
from unittest import TestCase, skip
from base64 import b64decode, b64encode
from time import time
import struct
import pickle
from hbase.utils import b64_encoder
from hbase.admin import HBaseAdmin
from hbase.rest_client import HBaseRESTClient
from hbase.put import Put
from hbase.get import Get
from hbase.delete import Delete


class TestDelete(TestCase):
    def setUp(self):
        self.client = HBaseRESTClient(hosts_list=["http://localhost:8080"])
        self.admin = HBaseAdmin(client=self.client)
        self.put = Put(client=self.client)
        self.get = Get(client=self.client)
        self.delete = Delete(client=self.client)

    def tearDown(self):
        self.client.session.close()

    def test_delete(self):
        self.admin.table_create_or_update(
            table_name="test_tbl", params_list=[{"name": "cf"}, {"name": "cf2"}]
        )

        self.put.put("test_tbl", "samir@example.com", {"cf:gender": "F"})
        self.put.put(
            "test_tbl", "samir@example.com", {"cf:countries": ["US", "BA", "DE", "SE"]}
        )
        self.put.put("test_tbl", "samir@example.com", {"cf:id": 13455})
        self.put.put("test_tbl", "samir@example.com", {"cf2:name": "samir"})

        self.delete.delete("test_tbl", "samir@example.com", "cf:gender")

        # Assert column not found after delete
        result = self.get.get("test_tbl", "samir@example.com", "cf:gender")
        assert result["status_code"] == 404

        # Delete column family
        self.delete.delete("test_tbl", "samir@example.com", "cf2")
        result = self.get.get("test_tbl", "samir@example.com", "cf2:name")
        assert result["status_code"] == 404

        # Delete row
        print(self.delete.delete("test_tbl", "samir@example.com"))
        result = self.get.get("test_tbl", "samir@example.com")
        assert result["status_code"] == 404



from hbase.utils import b64_encoder


class Put(object):
    def __init__(self, client):
        self.client = client

    @staticmethod
    def _build_url_suffix(tbl_name):
        return f"/{tbl_name}/fakerow"

    @staticmethod
    def _build_payload(row_key, cell_map, timestamp=None):
        cf_value_list = []
        for key, value in cell_map.items():
            if timestamp is None:
                cf_value_list.append(
                    {"column": b64_encoder(key), "$": b64_encoder(value)}
                )
            else:
                cf_value_list.append(
                    {
                        "column": b64_encoder(key),
                        "timestamp": timestamp,
                        "$": b64_encoder(value),
                    }
                )

        return {"Row": [{"key": b64_encoder(row_key), "Cell": cf_value_list}]}

    def _build_bulk_payload(self, rows, timestamp=None):
        row_list = []
        for r in rows:
            row_key = r[0]
            cell_map = r[1]
            if timestamp is None:
                row_list.append(self._build_payload(row_key, cell_map)["Row"][0])
            else:
                row_list.append(
                    self._build_payload(row_key, cell_map, timestamp=timestamp)["Row"][
                        0
                    ]
                )
        return {"Row": row_list}

    def put(self, tbl_name, row_key, cell_map, timestamp=None):
        return self.client.send_request(
            method="PUT",
            url_suffix=self._build_url_suffix(tbl_name),
            payload=self._build_payload(row_key, cell_map, timestamp),
        )

    def bulk(self, tbl_name, rows, timestamp=None):
        return self.client.send_request(
            method="PUT",
            url_suffix=self._build_url_suffix(tbl_name),
            payload=self._build_bulk_payload(rows, timestamp),
        )



import json
from hbase.utils import result_parser


class Get(object):
    def __init__(self, client):
        self.client = client

    @staticmethod
    def _build_url_suffix(
        tbl_name, row_key, column_family=None, timestamp=None, number_of_versions=None
    ):
        if timestamp is not None:
            return f"/{tbl_name}/{row_key}/{column_family}/{timestamp}"
        if number_of_versions is not None:
            return f"/{tbl_name}/{row_key}/{column_family}/?v={number_of_versions}"
        if column_family is not None:
            return f"/{tbl_name}/{row_key}/{column_family}"
        return f"/{tbl_name}/{row_key}"

    def get(
        self,
        tbl_name,
        row_key,
        column_family=None,
        timestamp=None,
        number_of_versions=None,
    ):
        result = self.client.send_request(
            method="GET",
            url_suffix=self._build_url_suffix(
                tbl_name, row_key, column_family, timestamp, number_of_versions
            ),
        )
        if isinstance(result, dict):
            return result
        return result_parser(json.loads(result))



class HBaseAdmin(object):
    def __init__(self, client):
        self.client = client

    def cluster_status(self):
        return self.client.send_request(method="GET", url_suffix=f"/status/cluster")

    def cluster_version(self):
        return self.client.send_request(method="GET", url_suffix=f"/version/cluster")

    def namespaces(self):
        return self.client.send_request(method="GET", url_suffix=f"/namespaces")

    def namespace(self, namespace_name):
        return self.client.send_request(
            method="GET", url_suffix=f"/namespaces/{namespace_name}"
        )

    def namespace_create(self, namespace_name):
        return self.client.send_request(
            method="POST", url_suffix=f"/namespaces/{namespace_name}"
        )

    def namespace_delete(self, namespace_name):
        return self.client.send_request(
            method="DELETE", url_suffix=f"/namespaces/{namespace_name}"
        )

    def delete_namespace(self):
        pass

    def tables(self, namespace="default"):
        return self.client.send_request(
            method="GET", url_suffix=f"/namespaces/{namespace}/tables"
        )

    def table_regions(self, table_name):
        return self.client.send_request(
            method="GET", url_suffix=f"/{table_name}/regions"
        )

    def table_schema(self, table_name):
        return self.client.send_request(
            method="GET", url_suffix=f"/{table_name}/schema"
        )

    def table_create_or_update(self, table_name, params_list):
        """
        Create table if not exist. Otherwise update table schema
        :param table_name:
        :param params_list: [{
                "name": "cfInfo",
                "BLOCKSIZE": "65536"
            }
            ]
        :return:
        """
        payload = {"name": table_name, "ColumnSchema": params_list}
        return self.client.send_request(
            method="PUT", url_suffix=f"/{table_name}/schema", payload=payload
        )

    def table_delete(self, table_name):
        return self.client.send_request(
            method="DELETE", url_suffix=f"/{table_name}/schema"
        )



import array
import pickle
from base64 import b64encode, b64decode
import struct

"""
HBase bytes int -> struct.unpack(">q", value) 8 bytes
HBase bytes float > struct.unpack(">d", value) 8 bytes
"""


def b64_encoder(value):
    value_type = type(value)
    if value_type in [bytes, bytearray, array.array]:
        return b64encode(value).decode("utf-8")
    elif value_type == str:
        return b64encode(value.encode("utf-8")).decode("utf-8")
    elif value_type == int:
        return b64encode(struct.pack(">q", value)).decode("utf-8")
    elif value_type == float:
        return b64encode(struct.pack(">d", value)).decode("utf-8")
    else:
        return b64encode(pickle.dumps(value)).decode("utf-8")


def result_parser(json_result):
    row = json_result["Row"]
    result = []
    key = {}
    cells = []
    for k in row:
        key["key"] = b64decode(k["key"])
        cell = k["Cell"]
        for i in cell:
            cells.append(
                {
                    "column": b64decode(i["column"]),
                    "timestamp": i["timestamp"],
                    "$": b64decode(i["$"]),
                }
            )
        key["cell"] = cells
        result.append(key)
        cells = []
        key = {}
    return {"row": result}


class Bytes(object):
    @staticmethod
    def to_int(bytes_value):
        return struct.unpack(">q", bytes_value)[0]

    @staticmethod
    def to_float(bytes_value):
        return struct.unpack(">d", bytes_value)[0]

    @staticmethod
    def to_string(bytes_values):
        return bytes_values.decode("utf-8")



import xml.etree.ElementTree as et
from hbase.utils import b64_encoder


# {"type":"SkipFilter","filters":[{"op":"NOT_EQUAL","type":"QualifierFilter","comparator":{"value":"dGVzdFF1YWxpZmllck9uZS0y","type":"BinaryComparator"}}]}
# {"op":"MUST_PASS_ALL","type":"FilterList","filters":[{"op":"EQUAL","type":"RowFilter","comparator":{"value":".+-2","type":"RegexStringComparator"}},{"op":"EQUAL","type":"QualifierFilter","comparator":{"value":".+-2","type":"RegexStringComparator"}},{"op":"EQUAL","type":"ValueFilter","comparator":{"value":"one","type":"SubstringComparator"}}]}
# {"op":"MUST_PASS_ONE","type":"FilterList","filters":[{"op":"EQUAL","type":"RowFilter","comparator":{"value":".+Two.+","type":"RegexStringComparator"}},{"op":"EQUAL","type":"QualifierFilter","comparator":{"value":".+-2","type":"RegexStringComparator"}},{"op":"EQUAL","type":"ValueFilter","comparator":{"value":"one","type":"SubstringComparator"}}]}


def build_base_xml(
    batch=1000,
    startRow="",
    endRow="",
    startTime="",
    endTime="",
    maxVersions=1,
    column=None,
):
    default = {
        "batch": str(batch),
        "startRow": b64_encoder(startRow) if startRow != "" else "",
        "endRow": b64_encoder(endRow) if endRow != "" else "",
        "startTime": str(startTime),
        "endTime": str(endTime),
        "maxVersions": str(maxVersions),
    }
    for k in list(default.keys()):
        if default[k] in [None, "None", ""]:
            del default[k]

    xml = et.Element("Scanner", attrib=default)

    if isinstance(column, list) and len(column) > 0:
        for c in column:
            column = et.SubElement(xml, "column")
            column.text = b64_encoder(c)
    return xml


def build_base_scanner(
    batch=1000,
    startRow="",
    endRow="",
    startTime="",
    endTime="",
    maxVersions=1,
    column=None,
):
    xml = build_base_xml(
        batch, startRow, endRow, startTime, endTime, maxVersions, column
    )
    return et.tostring(xml).decode("utf-8")


def build_prefix_filter(
    row_perfix,
    batch=1000,
    startRow="",
    endRow="",
    startTime="",
    endTime="",
    maxVersions=1,
    column=None,
):
    xml = build_base_xml(
        batch, startRow, endRow, startTime, endTime, maxVersions, column
    )

    filter = et.SubElement(xml, "filter")
    filter.text = '{"type":"PrefixFilter", "value":"%s"}' % (b64_encoder(row_perfix))
    return et.tostring(xml).decode("utf-8")


def build_row_filter(
    row_value,
    operation,
    comparator="binary",
    batch=1000,
    startRow="",
    endRow="",
    startTime="",
    endTime="",
    maxVersions=1,
    column=None,
):
    # {"op": "LESS", "type": "RowFilter", "comparator": {"value": "dGVzdFJvd09uZS0y", "type": "BinaryComparator"}}
    xml = build_base_xml(
        batch, startRow, endRow, startTime, endTime, maxVersions, column
    )

    filter = et.SubElement(xml, "filter")
    if comparator == "binary":
        filter.text = (
            '{"op":"%s","type":"RowFilter", "comparator": {"value": "%s", "type": "BinaryComparator"} }'
            % (operation, b64_encoder(row_value))
        )
    else:
        filter.text = (
            '{"op":"%s","type":"RowFilter", "comparator": {"value": "%s", "type": "RegexStringComparator"} }'
            % (operation, row_value)
        )

    return et.tostring(xml).decode("utf-8")


def build_value_filter(
    value,
    operation,
    comparator="binary",
    batch=1000,
    startRow="",
    endRow="",
    startTime="",
    endTime="",
    maxVersions=1,
    column=None,
):
    # {"op":"EQUAL","type":"ValueFilter","comparator":{"value":"dGVzdFZhbHVlT25l","type":"BinaryComparator"}}

    xml = build_base_xml(
        batch, startRow, endRow, startTime, endTime, maxVersions, column
    )
    filter = et.SubElement(xml, "filter")
    if comparator == "binary":
        filter.text = (
            '{"op":"%s","type":"ValueFilter", "comparator": {"value": "%s", "type": "BinaryComparator"} }'
            % (operation, b64_encoder(value))
        )
    else:
        filter.text = (
            '{"op":"%s","type":"ValueFilter", "comparator": {"value": "%s", "type": "SubstringComparator"}}'  # SubstringComparator
            % (operation, value)
        )

    return et.tostring(xml).decode("utf-8")


def build_single_column_value_filter(
    family,
    qualifier,
    value,
    operation,
    comparator="binary",
    batch=1000,
    startRow="",
    endRow="",
    startTime="",
    endTime="",
    maxVersions=1,
    column=None,
):
    """
    {
      "type": "SingleColumnValueFilter",
      "op": "EQUAL",
      "family": "ZmFtaWx5",
      "qualifier": "Y29sMQ\u003d\u003d",
      "latestVersion": true,
      "comparator": {
        "type": "BinaryComparator",
        "value": "MQ\u003d\u003d"
      }
    }
        :param operation:
        :param comparator:
        :param batch:
        :param type:
        :param startRow:
        :param endRow:
        :param startTime:
        :param endTime:
        :param maxVersions:
        :return:
    """
    xml = build_base_xml(
        batch, startRow, endRow, startTime, endTime, maxVersions, column
    )
    filter = et.SubElement(xml, "filter")
    if comparator == "binary":
        filter.text = '{"op":"%s","type":"SingleColumnValueFilter","family":"%s", "qualifier":"%s", "comparator": {"value": "%s", "type": "BinaryComparator"}}' % (
            operation,
            b64_encoder(family),
            b64_encoder(qualifier),
            b64_encoder(value),
        )

    else:
        filter.text = (
            '{"op":"%s","type":"SingleColumnValueFilter","family":"%s", "qualifier":"%s", "comparator": {"value": "%s", "type": "SubstringComparator"}}'
            % (operation, b64_encoder(family), b64_encoder(qualifier), value)
        )

    return et.tostring(xml).decode("utf-8")


def build_column_prefix_filter(
    column_prefix,
    batch=1000,
    startRow="",
    endRow="",
    startTime="",
    endTime="",
    maxVersions=1,
    column=None,
):
    xml = build_base_xml(
        batch, startRow, endRow, startTime, endTime, maxVersions, column
    )
    filter = et.SubElement(xml, "filter")
    filter.text = '{"type":"ColumnPrefixFilter", "value": "%s"}' % (
        b64_encoder(column_prefix)
    )
    return et.tostring(xml).decode("utf-8")



import json
import sys
import time
import datetime
from hbase.rest_client import HBaseRESTClient
from hbase.admin import HBaseAdmin
from hbase.put import Put
from hbase.get import Get
from hbase.scan import Scan
from hbase.utils import Bytes
from hbase.scan_filter_helper import (
    build_column_prefix_filter,
    build_value_filter,
    build_base_scanner,
)

"""
HBase scheme design example for storing user messages
cfmm:thread_id:message_id -> subject + body

"""


# Modes BaseModel, User, Message


class BaseModel(object):
    def __init__(self, data):

        self.defaults = {"_raw_json": None}
        setattr(self, "_raw_json", data)

    def __str__(self):
        return self.asJsonString()

    def asJsonString(self):
        return json.dumps(
            self.asDict(), sort_keys=True, separators=(",", ":"), indent=4
        )

    def asDict(self):

        data = {}

        for (key, value) in self._raw_json.items():
            if isinstance(self._raw_json.get(key, None), (list, tuple, set)):
                data[key] = list()
                for subobj in self._raw_json.get(key, None):
                    data[key].append(subobj)

            elif self._raw_json.get(key, None):
                data[key] = self._raw_json.get(key, None)

        return data


class User(BaseModel):
    def __init__(self, data):
        super(User, self).__init__(data=data)

        self.defaults = {"id": None, "name": None, "email": None}

        for (param, default) in self.defaults.items():
            setattr(self, param, data.get(param, default))


class Message(BaseModel):
    def __init__(self, data):
        super(Message, self).__init__(data=data)

        self.defaults = {
            "id": None,
            "subject": None,
            "body": None,
            "sender_email": None,
            "sender_name": None,
            "send_date": None
        }

        for (param, default) in self.defaults.items():
            setattr(self, param, data.get(param, default))

def get_datetime_from_id(id):
    tmstmp_from_id = sys.maxsize - int(id)
    return datetime.datetime.fromtimestamp(int(tmstmp_from_id/1000)).strftime('%Y-%m-%d %H:%M:%S')

def generate_id():
    tmstmp = round(time.time() * 1000)
    id = sys.maxsize - tmstmp
    return id

def create_inbox_table(client):
    admin = HBaseAdmin(client=client)
    admin.table_create_or_update(
        table_name="inbox",
        params_list=[{"name": "cfu"}, {"name": "cfmm"}],
    )

def register_user(client, user):
    put = Put(client=client)
    put.put(
        tbl_name="inbox",
        row_key=str(user.id),
        cell_map={"cfu:desc": f"{user.name}:{user.email}"},
    )


def get_user(client, user_id):
    get = Get(client=client)
    result = get.get(tbl_name="inbox", row_key=str(user_id), column_family="cfu:desc")
    name_email = Bytes.to_string(result["row"][0]["cell"][0]["$"])
    name_email_lst = name_email.split(":")
    user = User(
        data={"id": user_id, "name": name_email_lst[0], "email": name_email_lst[1]}
    )
    return user.asJsonString()


def create_user_message(client, user_id, msg):
    put = Put(client=client)
    put.put(
        tbl_name="inbox",
        row_key=str(user_id),
        cell_map={
            f"cfmm:{msg.id}:{msg.subject}:{msg.sender_email}:{msg.sender_name}": msg.body,
        },
    )

def get_msgs_from_result(msg_cells):
    messages = []
    for c in msg_cells:
        value = Bytes.to_string(c["$"])
        c_list = Bytes.to_string(c["column"]).split(":")
        if c_list[0] == "cfmm":
            messages.append(
                Message(
                    data={
                        "id": c_list[1],
                        "subject": c_list[2],
                        "body": value,
                        "send_date": get_datetime_from_id(c_list[1]),
                        "sender_email": c_list[3],
                        "sender_name": c_list[4],
                    }
                )
            )
    #m_sorted = sorted(messages, key=lambda k: k.send_date, reverse=True)
    return [m.asDict() for m in messages]

def get_user_messages(client, user_id):
    get = Get(client)
    data = get.get("inbox", str(user_id))
    user_id = Bytes.to_string(data["row"][0]["key"])
    columns = data["row"][0]["cell"]
    return get_msgs_from_result(msg_cells=columns)


def search_message(client, word, user_id):
    scan = Scan(client)
    filter = build_value_filter(
        startRow=str(user_id),
        endRow=str(user_id),
        value=word,
        column=['cfmm'],
        operation="EQUAL",
        comparator="substring",
    )
    print(filter)
    _, result = scan.scan(tbl_name="inbox", scanner_payload=filter)
    columns = result["row"][0]["cell"]
    scan.delete_scanner()
    return get_msgs_from_result(columns)


def get_message_by_id(client, message_id):
    scan = Scan(client)
    filter = build_column_prefix_filter(column_prefix=str(message_id), column=['cfmm'])
    _, result = scan.scan(tbl_name="inbox", scanner_payload=filter)
    columns = result["row"][0]["cell"]
    scan.delete_scanner()
    return get_msgs_from_result(msg_cells=columns)

def scan_mesages(client, batch):
    scn = Scan(client=client)
    filter = build_base_scanner(batch=batch)
    _, result = scn.scan('inbox', filter    )
    columns = result["row"][0]["cell"]
    return get_msgs_from_result(columns)



if __name__ == "__main__":

    client = HBaseRESTClient(hosts_list=["http://localhost:8080"])

    test_user = User(
        data={"id": 245, "name": "Samir Ahmic", "email": "samir@example.com"}
    )

    # test_message1 = Message(
    #     data={
    #         "id": generate_id(),
    #         "subject": "test subject",
    #         "body": "some message body",
    #         "sender_email": "test@example.com",
    #         "sender_name": "Jon Doe",
    #         "send_date": '2021-11-07'
    #     }
    # )
    # time.sleep(10)
    # test_message2 = Message(
    #     data={
    #         "id": generate_id(),
    #         "subject": "meeting notes",
    #         "body": "Last Monday ",
    #         "sender_email": "test2@example.com",
    #         "sender_name": "Susan Brown",
    #         "send_date": '2021-11-08'
    #     }
    # )
    # time.sleep(10)
    #
    # test_message3 = Message(
    #     data={
    #         "id": generate_id(),
    #         "subject": "Subscriptions execution plan",
    #         "body": "Hi, Samir. Here is plan for Apple subscriptions",
    #         "sender_email": "test3@example.com",
    #         "sender_name": "Bugs Bunny",
    #         "send_date": '2021-11-09',
    #     }
    # )
    # time.sleep(10)
    # test_message4 = Message(
    #     data={
    #         "id": generate_id(),
    #         "subject": "The Argentinian users case",
    #         "body": "The Argentinian bank account number field has been updated to allow for 22 digits.",
    #         "sender_email": "test4@example.com",
    #         "sender_name": "Jessica Doil",
    #         "send_date": '2021-11-10',
    #     }
    # )
    # time.sleep(10)
    #
    # test_message5 = Message(
    #     data={
    #         "id": generate_id(),
    #         "subject": "New DSPs, post launch",
    #         "body": "Here is new schema for post launch marketing",
    #         "sender_email": "test5@example.com",
    #         "sender_name": "Hanna Marou",
    #         "send_date": '2021-11-11',
    #     }
    # )
    #
    # time.sleep(10)
    #
    # test_message6 = Message(
    #     data={
    #         "id": generate_id(),
    #         "subject": "Using HFileLink files in split as default",
    #         "body": "Hi,I proposed an issue to use HFileLink file when splitting in",
    #         "sender_email": "test6@example.com",
    #         "sender_name": "Erik Ljunkvist",
    #         "send_date": '2021-11-12',
    #     }
    # )
    #
    # create_inbox_table(client)
    # register_user(client, test_user)
    # create_user_message(client, test_user.id, test_message1)
    # create_user_message(client, test_user.id, test_message2)
    # create_user_message(client, test_user.id, test_message3)
    # create_user_message(client, test_user.id, test_message4)
    # create_user_message(client, test_user.id, test_message5)
    # create_user_message(client, test_user.id, test_message6)


    print("\n--------- Get user ----------")
    print(get_user(client, test_user.id))
    # print(get_user_messages(client, test_user.id))


    print("\n----- Search message for containing word in message body------")
    print(search_message(client, "hfile", 245))

    print("----------- All results ---------")

    print(get_user_messages(client, 245))

    print("\n--- Get 5 last messages ----")
    print(scan_mesages(client, 5))






import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hbase-rest-py",
    version="1.0.0",
    author="Samir Ahmic",
    author_email="ahmic.samir@gmail.com",
    description="HBase client based on HBase REST",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/samirMoP/hbase-rest-py",
    project_urls={
        "Bug Tracker": "https://github.com/samirMoP/hbase-rest-py/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires = ['requests', 'requests-toolbelt'],
    tests_require = ['mock'],
    packages = setuptools.find_packages(exclude = ('tests', 'doc')),
    python_requires=">=3.6",
)


import json
import random
from requests import Session


class HBaseRESTClient(object):
    def __init__(self, hosts_list):
        self.hosts = hosts_list
        self.session = Session()
        self.headers = {
            "accept": "application/json",
            "content-type": "application/json",
        }
        self.timeout = 10

    def get_hbase_host(self):
        return random.choice(self.hosts)

    def send_request(self, method, url_suffix, payload=None):
        response = None
        host = self.get_hbase_host()
        if method == "GET":
            response = self.session.get(
                url=f"{host}{url_suffix}",
                headers=self.headers,
                timeout=self.timeout,
            )
        if method == "PUT":
            response = self.session.put(
                url=f"{host}{url_suffix}",
                data=json.dumps(payload) if payload is not None else None,
                headers=self.headers,
                timeout=self.timeout,
            )
        if method == "POST":
            response = self.session.post(
                url=f"{host}{url_suffix}",
                data=json.dumps(payload) if payload is not None else None,
                headers=self.headers,
                timeout=self.timeout,
            )
        if method == "DELETE":
            response = self.session.delete(
                url=f"{host}{url_suffix}",
                headers=self.headers,
                timeout=self.timeout,
            )

        if response.status_code in [200, 201]:
            return response.content.decode("utf-8")
        else:
            return {
                "error": response.content.decode("utf-8"),
                "status_code": response.status_code,
            }
